#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>

#include "jc.h"


void JCFBOpenCalibCB(GtkWidget *w, GtkFileSelection *fs);
void JCFBSaveCalibCB(GtkWidget *w, GtkFileSelection *fs);
void JCFBCancelCB(GtkWidget *w, GtkFileSelection *fs);
void JCFBDestroyCB(GtkWidget *widget, GtkFileSelection *fs);

gint JCFBCreate(GtkWidget **fb, gint argc, gchar **argv);
void JCFBMapByCommand(GtkWidget **fb, gint command, jc_struct *jc);
void JCFBDestroy(GtkWidget **fb);


/*
 *	Centers toplevel window to root.
 */
void JCFBCenterToplevel(GtkWidget *fb)
{
        GtkWidget *w;
        GdkWindow *parent;
        gint width1, height1, width2, height2;


        w = fb;
        if(w == NULL)
            return;

        parent = gdk_window_get_parent(w->window);
        if(parent != NULL)
        {
            gdk_window_get_size(w->window, &width1, &height1);
            gdk_window_get_size(parent, &width2, &height2);

            gtk_widget_set_uposition(
                w,
                ((int)width2 / 2) - ((int)width1 / 2),
                ((int)height2 / 2) - ((int)height1 / 2)
            );
        }

        return;
}


/*
 *	File browser open calibration callback.
 */
void JCFBOpenCalibCB(GtkWidget *w, GtkFileSelection *fs)
{
        if(fs == NULL)
            return;

        /* Unmap file browser. */
        gtk_widget_hide(GTK_WIDGET(fs));

	/* Perform open calibration file procedure. */
	JCDoOpenCalibration(
	    &js_calib,
	    gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs))
	);

	return;
}

/*
 *	File browser save calibration callback.
 */
void JCFBSaveCalibCB(GtkWidget *w, GtkFileSelection *fs)
{
        if(fs == NULL)
            return;

        /* Unmap file browser. */
        gtk_widget_hide(GTK_WIDGET(fs));

        /* Perform save calibration file procedure. */
        JCDoSaveCalibration(
            &js_calib,
            gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs))
        );

	return;
}

/*
 *	File browser cancel callback.
 */
void JCFBCancelCB(GtkWidget *w, GtkFileSelection *fs)
{
	if(fs == NULL)
	    return;

	/* Just unmap it. */
	gtk_widget_hide(GTK_WIDGET(fs));

	return;
}


/*
 *	File browser destroy callback, this just unmaps it.
 */
void JCFBDestroyCB(GtkWidget *widget, GtkFileSelection *fs)
{
        if(fs == NULL)
            return;

        /* Just unmap it. */
        gtk_widget_hide(GTK_WIDGET(fs));

	return;
}

/*
 *	Creates the file browser.
 */
gint JCFBCreate(GtkWidget **fb, gint argc, gchar **argv)
{
	if(fb == NULL)
	    return(-1);

	/* Create the file browser. */
	*fb = gtk_file_selection_new("File selection");
	if(*fb == NULL)
	    return(-1);

	gtk_widget_realize(*fb);

	/* Set destroy callback. */
	gtk_signal_connect(
	    GTK_OBJECT(*fb), "destroy",
	    GTK_SIGNAL_FUNC(JCFBDestroyCB),
	    *fb
	);

	/* Set callback for the cancel button. */
	gtk_signal_connect(
	    GTK_OBJECT(GTK_FILE_SELECTION(*fb)->cancel_button),
	    "clicked",
	    GTK_SIGNAL_FUNC(JCFBCancelCB),
            *fb
	);

     

	return(0);
}

/*
 *	Maps the file browser and sets it up according to the
 *	given command which can be any of JC_CMD_*.
 */
void JCFBMapByCommand(GtkWidget **fb, gint command, jc_struct *jc)
{
	if((fb == NULL) ||
	   (jc == NULL)
	)
	    return;
	if(*fb == NULL)
	    return;

	switch(command)
	{
	  case JC_CMD_NONE:
	    break;

	  /* Open calibration file. */
	  case JC_CMD_OPEN_CALIB:
            /* Set callback for the OK button. */
	    gtk_signal_connect(
		GTK_OBJECT(GTK_FILE_SELECTION(*fb)->ok_button),
		"clicked",
		GTK_SIGNAL_FUNC(JCFBOpenCalibCB),
		*fb
	    );
            gtk_file_selection_complete(
		GTK_FILE_SELECTION(*fb),
		".*"
	    );
	    gtk_file_selection_set_filename(
		GTK_FILE_SELECTION(*fb),
                ((jc->calib_file == NULL) ? JC_DEF_JS_CALIB : jc->calib_file)
	    );
	    gtk_window_set_title(
		GTK_WINDOW(*fb),
		"Open Calibration File"
	    );
	    JCFBCenterToplevel(*fb);
	    gtk_widget_show(*fb);
	    break;

	  /* Save as calibration file. */
	  case JC_CMD_SAVEAS_CALIB:
            /* Set callback for the OK button. */
            gtk_signal_connect(
                GTK_OBJECT(GTK_FILE_SELECTION(*fb)->ok_button),
                "clicked",
                GTK_SIGNAL_FUNC(JCFBSaveCalibCB),
                *fb
            );
            gtk_file_selection_complete(
                GTK_FILE_SELECTION(*fb),
                ".*"  
            );
            gtk_file_selection_set_filename(
                GTK_FILE_SELECTION(*fb),
                ((jc->calib_file == NULL) ? JC_DEF_JS_CALIB : jc->calib_file)
            );
            gtk_window_set_title(
                GTK_WINDOW(*fb),
                "Save Calibration File"
            );
            JCFBCenterToplevel(*fb);
            gtk_widget_show(*fb);
            break;
	}

	return;
}

/*
 *	Destroys the file browser.
 */
void JCFBDestroy(GtkWidget **fb)
{


	if(fb == NULL)
	    return;
        if(*fb == NULL)
            return;




	*fb = NULL;

	return;
}
