#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "../include/cfgfmt.h"
#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/fio.h"
#include "../include/disk.h"

#include "cdialog.h"
#include "jc.h"


#ifdef __linux__
static void JCWriteAxisCalibBlock(
	FILE *fp, js_data_struct *jsd, gint axis_num
);
static void JCWriteJSCalibBlock(FILE *fp, js_data_struct *jsd);
#endif	/* __linux__ */

gint JCDoSaveCalibration(jc_struct *jc, const gchar *path);


#ifdef __linux__
/*
 *	Writes an axis calibration block for the specified axis_num
 *	on the jsd structure to the stream pointed to by fp.
 */
static void JCWriteAxisCalibBlock(
	FILE *fp, js_data_struct *jsd, gint axis_num
)
{
	js_axis_struct *axis_ptr;


	if((fp == NULL) || (jsd == NULL))
	    return;

	if(JSIsAxisAllocated(jsd, axis_num))
	    axis_ptr = jsd->axis[axis_num];
	else
	    return;

	/* Begin writing axis calibration block. */
        fprintf(
            fp,
            "    BeginAxis = %i\n",
            axis_num
        );
        fprintf(
            fp,
            "        Minimum = %i\n",
            axis_ptr->min
        );
        fprintf(
            fp,
            "        Center = %i\n",
            axis_ptr->cen
        );
        fprintf(
            fp,
            "        Maximum = %i\n",
            axis_ptr->max
        );
        fprintf(
            fp,
            "        NullZone = %i\n",
            axis_ptr->nz
        );
	/* Always enable tolorance. */
	fprintf(
	    fp,
	    "        Tolorance = %i\n",
	    axis_ptr->tolorance
	);
	if(axis_ptr->flags & JSAxisFlagFlipped)
            fprintf(
                fp,
                "        Flip\n"
            );
	if(axis_ptr->flags & JSAxisFlagIsHat)
            fprintf( 
                fp,
                "        IsHat\n"
            );
        fprintf(fp, "    EndAxis\n");

	return;
}
#endif  /* __linux__ */

#ifdef __linux__
/*
 *	Writes a joystick calibration block for the joystick specified
 *	by jsd to the stream pointed to by fp.
 */
static void JCWriteJSCalibBlock(FILE *fp, js_data_struct *jsd)
{
	gint axis_num;

	if((fp == NULL) || (jsd == NULL))
	    return;

	/* Begin writing joystick calibration block. */

	/* Begin joystick configuration block statement. */
        fprintf(fp, "BeginJoystick = %s\n", jsd->device_name);

	/* Write each axis. */
        for(axis_num = 0; axis_num < jsd->total_axises; axis_num++)
	    JCWriteAxisCalibBlock(fp, jsd, axis_num);

/* Other configuration that should be written needs to be done here. */

	/* End joystick configuration block statement. */
        fprintf(fp, "EndJoystick\n");

	return;
}
#endif	/* __linux__ */


/*
 *	Saves calibration values in the jsd structure on the specified
 *	jc to the given calibration file specified by path.
 *
 *	If path is NULL then the default calibration file name in the
 *	user's home directory will be used.
 */
gint JCDoSaveCalibration(jc_struct *jc, const gchar *path)
{
	gint i, n;
	gchar *strptr;
        gboolean wrote_our_js = FALSE;

	gint strc = 0;
	gchar **strv = NULL;
	FILE *fp;

	js_data_struct *our_jsd, *jsd_ptr;

	js_data_struct **jsd;
	gint total_jsds;

        gchar lcalib[PATH_MAX + NAME_MAX];
        gchar ldevice[PATH_MAX + NAME_MAX];


	if(jc == NULL)
	    return(-1);

	/* Get pointer to `our' jsd, the one specified on the jc
	 * structure.
	 */
	our_jsd = &jc->jsd;

	/* Calibration file name specified? */
	if(path == NULL)
	{
	    /* Calibration file name not specified, so use default
	     * calibration file name assumed located in the user's home
	     * directory.
	     */
	    strptr = getenv("HOME");
	    strptr = PrefixPaths(
		(strptr == NULL) ? "/" : strptr,
		JC_DEF_JS_CALIB
	    );
	    strncpy(
		lcalib,
		(strptr == NULL) ? JC_DEF_JS_CALIB : strptr,
		PATH_MAX + NAME_MAX
	    );
	}
	else
	{
	    /* Calibration file is specified, copy it to the
	     * local calibration file name string.
	     */
            strncpy(lcalib, path, PATH_MAX + NAME_MAX);
	}
	lcalib[PATH_MAX + NAME_MAX - 1] = '\0';

	/* Get the device name for the jsd structure on the given jc
	 * structure. If no device name is specified then use the
	 * default device name.
	 */
	strptr = our_jsd->device_name;
	strncpy(
	    ldevice,
	    (strptr == NULL) ? JC_DEF_JS_DEVICE : strptr,
	    PATH_MAX + NAME_MAX
	);
	ldevice[PATH_MAX + NAME_MAX - 1] = '\0';


#ifdef __linux__
	/* Get listing of devices in calibration file for Linux. */
	strv = JSLoadDeviceNamesLinux(&strc, lcalib);

	/* Allocate and set values for each joystick device, going
	 * through each device name loaded from the calibration
	 * file.
	 */
	for(i = 0, jsd = NULL, total_jsds = 0;
            i < strc;
            i++
	)
	{
	    strptr = strv[i];
	    if(strptr == NULL)
		continue;

	    /* Allocate more pointers for the local jsd structures
	     * list.
	     */
	    n = total_jsds;
	    total_jsds++;
	    jsd = (js_data_struct **)realloc(
		jsd,
		total_jsds * sizeof(js_data_struct *)
	    );
	    if(jsd == NULL)
	    {
		total_jsds = 0;
		continue;
	    }

	    /* Allocate a new jsd structure in our local list. */
	    jsd[n] = jsd_ptr = (js_data_struct *)calloc(
		1,
		sizeof(js_data_struct)
	    );
	    if(jsd_ptr == NULL)
		continue;

	    /* Set up jsd values for our newly allocated local jsd
	     * structure.
	     */
	    jsd_ptr->fd = -1;
	    jsd_ptr->device_name = StringCopyAlloc(strptr);
	    jsd_ptr->calibration_file = StringCopyAlloc(lcalib);

	    /* Call joystick wrapper library loading function for
	     * Linux. Load the calibration values into our local
	     * jsd structure.
	     */
	    JSLoadCalibrationLinux(jsd_ptr);
	}

	/* We should now have a local list of jsd structures which
	 * contain all values from the specified joysticks in the
	 * calibration file.
	 */

	/* Free device names, they are no longer needed. */
	StringFreeArray(strv, strc);
	strv = NULL;
	strc = 0;


	/* Open calibration file for writing. */
	fp = fopen(lcalib, "w");
	if(fp != NULL)
	{
	    /* Write header. */
	    fprintf(fp,
"# Joystick calibration file.\n\
# Generated by %s version %s.\n\
#\n",
		PROG_NAME, PROG_VERSION
	    );

	    /* Write each joystick device block. */
	    for(i = 0; i < total_jsds; i++)
	    {
		jsd_ptr = jsd[i];
		if(jsd_ptr == NULL)
		    continue;

		/* Skip if device name is not specified, because we
		 * need the device name specified on the jsd to check
		 * if the device name matches that of `our' jsd.
		 */
		if(jsd_ptr->device_name == NULL)
		    continue;

		/* Check if joystick device values we are about to
		 * write is our currently opened joystick device.
		 */
		if(strcmp(jsd_ptr->device_name, ldevice))
		{
		    /* Calibration block for some other joystick,
		     * just write the calibration for that joystick
		     * to the calibration file as is.
		     */
		    JCWriteJSCalibBlock(fp, jsd_ptr);
		}
		else
		{
		    /* `Our' joystick, do not use the calibration for
		     * `our' joystick from the local jsd list, instead
		     * use the values for the jsd from the jc structure
		     * (since that is `our' joystick afterall.
		     */
		    JCWriteJSCalibBlock(fp, our_jsd);

		    /* Mark that we already wrote our joystick device. */
		    wrote_our_js = TRUE;
		}
	    }

	    /* All joystick calibration blocks from the local jsd list
	     * should now all be written to the file. But we need to
	     * double check if `our' joystick calibration information
	     * was written or not. Since this might occure if our joystick
	     * was not specified at all in the calibration file.
	     */
	    if(!wrote_our_js)
	    {
		/* Looks like our joystick calibration values were not
		 * written, so write it now.
		 */
                JCWriteJSCalibBlock(fp, our_jsd);

                /* Mark that we have now written values for our joystick
		 * device.
		 */
                wrote_our_js = TRUE;
	    }

	    /* Close the file. */
	    fclose(fp);
	    fp = NULL;
	}


	/* Free our local list of jsd structures. */
	for(i = 0; i < total_jsds; i++)
	{
	    jsd_ptr = jsd[i];
	    if(jsd_ptr == NULL)
		continue;

	    /* Calling JSClose() will deallocate all resources and
	     * substructures in the jsd, note that since we set the
	     * member fd in all the local jsd's they won't be closed
	     * since they were never opened.
	     */
	    JSClose(jsd_ptr);

	    /* We need to free the jsd structure itself. */
	    free(jsd_ptr);
	}
	/* Free the local jsd list pointer array. */
	free(jsd);
	jsd = NULL;
	total_jsds = 0;

#endif  /* __linux__ */


	/* Check if we successfully saved `our' joystick. */
	if(wrote_our_js)
	{
	    /* Mark calibration values saved. */
	    if(jc->has_changes)
		jc->has_changes = FALSE;

	    StatusBarSetMesg(
                &jc->status_bar,
                "Calibration saved"
            );
	}
	else
	{
            gchar text[256 + (2 * PATH_MAX + NAME_MAX)];


            /* Save failed. */
            StatusBarSetMesg(&jc->status_bar, NULL);

            sprintf(
                text,
"Unable to save calibration for joystick:\n\n\
    %s\n\n\
To calibration file:\n\n\
    %s",
		ldevice,
                lcalib
            );
	    CDialogSetTransientFor(jc->toplevel);
            CDialogGetResponse(
"Failed to save calibration",
text,
"The joystick device calibration could not be saved\n\
due to an error. Please verify that the specified\n\
location for the calibration file to be saved is writeable.",
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
	    CDialogSetTransientFor(NULL);
	}

	return(0);
}
