#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include "../include/cfgfmt.h"
#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/disk.h"
#include "../include/fio.h"

#include "../include/jsw.h"


/* Public functions. */
int JSIsInit(js_data_struct *jsd);
js_attribute_struct *JSGetAttributesList(
        int *total, const char *calibration
);
void JSFreeAttributesList(js_attribute_struct *list, int total);
unsigned int JSDriverVersion(js_data_struct *jsd);
int JSDriverQueryVersion(
	js_data_struct *jsd,
	int *major_rtn, int *minor_rtn, int *release_rtn
);
int JSIsAxisAllocated(js_data_struct *jsd, int n);
int JSIsButtonAllocated(js_data_struct *jsd, int n);

double JSGetAxisCoeff(js_data_struct *jsd, int n);
double JSGetAxisCoeffNZ(js_data_struct *jsd, int n);
int JSGetButtonState(js_data_struct *jsd, int n);

void JSResetAllAxisTolorance(js_data_struct *jsd);
#ifdef __linux__
int JSLoadCalibrationLinux(js_data_struct *jsd);
char **JSLoadDeviceNamesLinux(
	int *total, const char *calibration
);
#endif	/* __linux__ */
int JSInit(
	js_data_struct *jsd,
	const char *device,
	const char *calibration,
	unsigned int flags
);
int JSUpdate(js_data_struct *jsd);
void JSClose(js_data_struct *jsd);



/*
 *	Checks if the joystick is initialized.
 */
int JSIsInit(js_data_struct *jsd)
{
	if(jsd == NULL)
	    return(0);
	else if((jsd->flags & JSFlagIsInit) &&
                (jsd->fd > -1)
	)
	    return(1);
	else
	    return(0);
}


/*
 *      Fetches attributes for all joysticks accessable (configured
 *      or not) on the system. Does not matter if the joystick is
 *      already opened or not. Returned values need to be free()'ed
 *      with a call to JSFreeAttributeList().
 *
 *      If the given calibration file calibration is NULL then some values
 *	may not be set in the attrib structure (ie is_configured and name).
 */
js_attribute_struct *JSGetAttributesList(
	int *total, const char *calibration
)
{
	int n, cur_total = 0;
	js_attribute_struct *attrib = NULL, *attrib_ptr;


	if(total != NULL)
	    (*total) = 0;

/* Allocates a new js_attribute_struct to the attrib list and
 * sets attrib_ptr to point to the new structure. Note that
 * attrib_ptr can be NULL on error.
 */
#define DO_ALLOCATE	\
{ \
 n = cur_total; \
 cur_total++; \
 attrib = (js_attribute_struct *)realloc( \
  attrib, cur_total * sizeof(js_attribute_struct) \
 ); \
 if(attrib == NULL) \
 { \
  cur_total = 0; \
  attrib_ptr = NULL; \
 } \
 else \
 { \
  attrib_ptr = &attrib[n]; \
  memset(attrib_ptr, 0x00, sizeof(js_attribute_struct)); \
 } \
}

#ifdef __linux__
	/* Begin fetching list of device attributes for Linux. */
	if(1)
	{
	    int i, j, fd, total_device_names;
	    char *strptr;
	    char **device_name;
	    char tmp_path[PATH_MAX + NAME_MAX];
	    struct stat stat_buf;


	    /* Begin searching for each device untill one is missing,
	     * look in the "/dev/js#" paths where # is a number.
	     */
	    for(i = 0; 1; i++)
	    {
		sprintf(tmp_path, "/dev/js%i", i);

		/* Not exist? */
		if(stat(tmp_path, &stat_buf))
		     break;

		/* Allocate a new attrib structure, set values to it. */
		DO_ALLOCATE
		if(attrib_ptr != NULL)
		{
		    /* Record device name. */
		    attrib_ptr->device_name = strdup(tmp_path);

		    /* See if we can open it. */
		    fd = open(tmp_path, O_RDONLY);
		    if(fd < 0)
		    {
			/* Could not open it, handle errno. */
			switch(errno)
			{
			  case ENODEV: case ENFILE:
			    attrib_ptr->not_accessable = 1;
			    break;

			  default:
			    /* All else assume is in use. */
			    attrib_ptr->is_in_use = 1;
			    break;
			}
		    }
		    else
		    {
			/* Opened it just fine, now close it. */
			close(fd);
		    }
		    /* From here fd is invalid. */
		}
	    }

            /* Repeat the above, except for USB devices, begin searching
	     * in the "/dev/input/js#" paths where # is a number.
             */
            for(i = 0; 1; i++)
            {
                sprintf(tmp_path, "/dev/input/js%i", i);

		/* Not exist? */
                if(stat(tmp_path, &stat_buf))
                     break;

                /* Allocate a new attrib structure, set values to it. */
                DO_ALLOCATE
                if(attrib_ptr != NULL)
                {
                    /* Record device name. */
                    attrib_ptr->device_name = strdup(tmp_path);

                    /* See if we can open it. */
                    fd = open(tmp_path, O_RDONLY);
                    if(fd < 0)
                    {
                        /* Could not open it, handle errno. */
                        switch(errno)
                        {
                          case ENODEV: case ENFILE:
                            attrib_ptr->not_accessable = 1;
                            break;

                          default:
                            /* All else assume is in use. */
                            attrib_ptr->is_in_use = 1;
                            break;
                        }
                    }
                    else
                    {
                        /* Opened it just fine, now close it. */
                        close(fd);
                    }
		    /* From here fd is invalid. */
                }
            }


	    /* Most attributes have been fetched and devices that exist
	     * have been checked for. Now scan through calibration file
	     * (if calibration is not NULL) and see which devices are
	     * configured properly.
	     */
	    device_name = JSLoadDeviceNamesLinux(
		&total_device_names, calibration
	    );

	    /* Itterate through attrib list, check which ones match
	     * a name in the device names list.
	     */
	    for(i = 0; i < cur_total; i++)
	    {
		attrib_ptr = &attrib[i];
		if(attrib_ptr->device_name == NULL)
		    continue;

		/* Itterate through device names, see if a device
		 * name in the list matches this device's name.
		 */
		for(j = 0; j < total_device_names; j++)
		{
		    strptr = device_name[j];
		    if(strptr == NULL)
			continue;

		    if(strcmp(strptr, attrib_ptr->device_name))
		        continue;

		    /* This device is in the device names list, so
		     * mark it as being configured properly.
		     */
		    attrib_ptr->is_configured = 1;
		    break;
		}
	    }

	    /* Free loaded device names. */
	    StringFreeArray(device_name, total_device_names);
	}
#else
	/* Some other platform. */




#endif

#undef DO_ALLOCATE

	/* Update total returned joystick attributes structures. */
        if(total != NULL)
            (*total) = cur_total;
 
	return(attrib);
}

/*
 *      Frees a list of js_attribute_structs and their substructures.
 */
void JSFreeAttributesList(js_attribute_struct *list, int total)
{
	int i;
	js_attribute_struct *ptr;


	for(i = 0; i < total; i++)
	{
	    ptr = &list[i];

	    free(ptr->name);
	    free(ptr->device_name);
	}
	free(list);
}

/*
 *      Returns the driver version value from the driver unparsed.
 *
 *	If you want the parsed one use JSDriverQueryVersion().
 *
 *	This requires that the calling function knows how to handle
 *	the returned version value.
 */
unsigned int JSDriverVersion(js_data_struct *jsd)
{
        if(jsd == NULL)
            return(0);
        else
            return(jsd->driver_version);
}

/*
 *	Returns the joystick driver version formatted, if any
 *	inputs are NULL then that value will not be set.
 *
 *	This function has some disadvantages compared to JSDriverVersion()
 *	as this function may leave out some additional pieces of
 *	information, if available (ie vendor code, if the implmentation
 *	supports it).
 *
 *	Returns 0 on failure and 1 on success.
 */
int JSDriverQueryVersion(
        js_data_struct *jsd, 
        int *major_rtn, int *minor_rtn, int *release_rtn 
)
{
	unsigned int ver_unparsed;

	/* Reset returns. */
	if(major_rtn != NULL)
	    (*major_rtn) = 0;
	if(minor_rtn != NULL)
	    (*minor_rtn) = 0;
	if(release_rtn != NULL)
	    (*release_rtn) = 0;

	if(jsd == NULL)
	    return(0);

#ifdef __linux__
	ver_unparsed = JSDriverVersion(jsd);

	if(major_rtn != NULL)
	    (*major_rtn) = (int)((ver_unparsed >> 16) & 0xff);
	if(minor_rtn != NULL)
	    (*minor_rtn) = (int)((ver_unparsed >> 8) & 0xff);
        if(release_rtn != NULL)
            (*release_rtn) = (int)(ver_unparsed & 0xff); 

	return(1);	/* Success. */
#else
	/* Platform unsupported, return failure. */
	return(0);
#endif
}

/*
 *	Checks if axis n is allocated on jsd.
 */
int JSIsAxisAllocated(js_data_struct *jsd, int n)
{
	if(jsd == NULL)
	    return(0);
	else if((n < 0) || (n >= jsd->total_axises))
	    return(0);
	else if(jsd->axis[n] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *      Checks if button n is allocated on jsd.
 */
int JSIsButtonAllocated(js_data_struct *jsd, int n)
{
        if(jsd == NULL)
            return(0);
        else if((n < 0) || (n >= jsd->total_buttons))
            return(0);
        else if(jsd->button[n] == NULL)
            return(0);
        else
            return(1);
}

/*
 *	Returns coefficient value from -1 to 1 for current
 *	axis position.
 */
double JSGetAxisCoeff(js_data_struct *jsd, int n)
{
	int x, dx, r;
	js_axis_struct *axis;

	if(JSIsAxisAllocated(jsd, n))
	{
	    /* Calculate axis position, all calculations in raw units. */

	    axis = jsd->axis[n];	/* Pointer to axis struct. */
	    x = axis->cur;		/* Current axis position. */
	    dx = x - axis->cen;		/* Delta from center. */

	    if(dx < 0)
	    {
		/* Negative position. */
		r = axis->min - axis->cen;	/* Negative range. */
		if(r < 0)
		    return(
		        (double)dx / (double)r *
			(double)((axis->flags & JSAxisFlagFlipped) ? 1 : -1)
		    );
		else
		    return(0.0);
	    }
	    else
	    {
		/* Positive position. */
                r = axis->max - axis->cen;	/* Positive range. */
                if(r > 0)
                    return(
			(double)dx / (double)r *
			(double)((axis->flags & JSAxisFlagFlipped) ? -1 : 1)
                    );
                else
                    return(0.0);
	    }
	}
	else
	{
	    return(0.0);
	}
}

/*
 *	Same as JSGetAxisCoefficient() except that it takes
 *	the nullzone into account.  Returning 0 if the position
 *	is within the nullzone.
 */
double JSGetAxisCoeffNZ(js_data_struct *jsd, int n)
{
        int x, dx, r, nz;
        js_axis_struct *axis;

        if(JSIsAxisAllocated(jsd, n))
        {
	    /* Calculate axis position, all calculations in raw units. */

	    axis = jsd->axis[n];	/* Pointer to axis struct. */
	    x = axis->cur;		/* Current axis position. */
	    dx = x - axis->cen;		/* Delta from center. */
	    nz = axis->nz;		/* Null zone absolute value. */

	    /* Null zone check. */
            if((dx < nz) && (dx > -nz))
		return(0.0);

            if(dx < 0)
            {
		/* Negative position. */
                r = axis->min - axis->cen;	/* Negative range. */
                if((r + nz) < 0)
                    return(
                        (double)(dx + nz) / (double)(r + nz) *
			(double)((axis->flags & JSAxisFlagFlipped) ? 1 : -1)
                    );
                else
                    return(0.0);
            }   
            else
            {
		/* Positive position. */
                r = axis->max - axis->cen;	/* Positive range. */
                if((r - nz) > 0)
                    return(
                        (double)(dx - nz) / (double)(r - nz) *
			(double)((axis->flags & JSAxisFlagFlipped) ? -1 : 1)
                    );
                else
                    return(0.0);
            }
        }
        else 
        {
	    /* Axis is not allocated. */
            return(0.0);
        }
}

/*
 *      Gets the button state of button n, one of JSButtonState*.
 */
int JSGetButtonState(js_data_struct *jsd, int n)
{
        if(JSIsButtonAllocated(jsd, n))
	    return(jsd->button[n]->state);
	else
	    return(JSButtonStateOff);
}


/*
 *	Syncs values for tolorance in error correction specified in the
 *	current axis structures to the low-level joystick driver.
 */
void JSResetAllAxisTolorance(js_data_struct *jsd)
{
	if(!JSIsInit(jsd))
	    return;

#ifdef __linux__
	if(jsd->total_axises > 0)
	{
	    int i;
	    js_axis_struct *axis_ptr;
            struct js_corr *corr;


            corr = (struct js_corr *)calloc(
                jsd->total_axises,
                sizeof(struct js_corr)
            );
            if(corr == NULL)
		return;

            for(i = 0; i < jsd->total_axises; i++)
            {
                axis_ptr = jsd->axis[i];
                if(axis_ptr == NULL)
                    continue;

                corr[i].type = JS_CORR_NONE;
                corr[i].prec = ((axis_ptr->flags & JSAxisFlagTolorance) ?
                    axis_ptr->tolorance : 0
                );
	    }

            if(ioctl(jsd->fd, JSIOCSCORR, corr))
                perror("setting correction");

            free(corr);
	}
#endif	/* __linux__ */


	return;
}


#ifdef __linux__
/*
 *	This function is internally called (as neede) by JSInit()
 *	external functions need not call this.
 *
 *	Loads calibration information from the calibration file
 *	specified in the jsd structure and loads calibration information
 *	into the jsd structure.
 */
int JSLoadCalibrationLinux(js_data_struct *jsd)
{
	int i, p;
	const char *this_device_name;
	char is_this_device = 0;
        char *strptr, *strptr2;
  
        FILE *fp;
  
        char parm[CFG_PARAMETER_MAX];
        char val[CFG_VALUE_MAX];
        int lines_read = 0;

	int axis_num, button_num;
	js_axis_struct *axis_ptr;
	js_button_struct *button_ptr;


	if(jsd == NULL)
	    return(-1);

	/* Device name must be specified. */
	this_device_name = (const char *)jsd->device_name;
	if(this_device_name == NULL)
	    return(-1);

	/* Calibration file must be specified. */
        if(jsd->calibration_file == NULL)
            return(-1);

	/* Open calibration file. */
        fp = fopen(jsd->calibration_file, "rb");
        if(fp == NULL)
	    return(-1);

/* Reads the next line from fp, first free()ing strptr and then allocating
 * the new line to strptr. Then uses strptr2 to parse the parameter and
 * value, the parsed values will be placed in the variables parm and val,
 * The value of lines_read will also be incremented.
 */
#define DO_GET_LINE	\
{ \
 /* Read next non-comment line. */ \
 free(strptr); \
 strptr = FReadNextLineAllocCount( \
  fp, UNIXCFG_COMMENT_CHAR, &lines_read \
 ); \
 if(strptr == NULL) \
  break; \
 \
 /* Fetch parameter. */ \
 strptr2 = StringCfgParseParm(strptr); \
 if(strptr2 == NULL) \
  continue; \
 strncpy(parm, strptr2, CFG_PARAMETER_MAX); \
 parm[CFG_PARAMETER_MAX - 1] = '\0'; \
 \
 /* Fetch value. */ \
 strptr2 = StringCfgParseValue(strptr); \
 if(strptr2 == NULL) \
  strptr2 = "0"; \
 strncpy(val, strptr2, CFG_VALUE_MAX); \
 val[CFG_VALUE_MAX - 1] = '\0'; \
}

	/* Begin reading file. */
	strptr = NULL;

        while(1)
        {
	    DO_GET_LINE

            /* BeginJoystick */
            if(!strcasecmp(parm, "BeginJoystick"))
            {
		/* Joystick device name must match the value for
		 * the BeginJoystick parameter, this is so that we
		 * know that this is a configuration block for
		 * this device (jsd->device_name).
		 */
		if(strcmp(val, this_device_name))
		    is_this_device = 0;
		else
		    is_this_device = 1;

		/* Enter loop to read and handle each line for this
		 * configuration block.
		 */
                while(1)
                {
		    DO_GET_LINE

                    /* BeginAxis */
                    if(!strcasecmp(parm, "BeginAxis") &&
                       is_this_device
                    )
                    {
			axis_ptr = NULL;
			axis_num = atoi(val);
			if(axis_num < 0)
			{
			    fprintf(
				stderr,
				"%s: Line %i: Invalid axis index %i.\n",
				jsd->calibration_file,
				lines_read,
				axis_num
			    );
			    axis_num = 0;	/* Index must be 0 or higher. */
			}

			/* Need to allocate more axises? */
			if(axis_num >= jsd->total_axises)
			{
			    p = jsd->total_axises;
			    jsd->total_axises = axis_num + 1;
			    jsd->axis = (js_axis_struct **)realloc(
				jsd->axis,
				jsd->total_axises * sizeof(js_axis_struct *)
			    );
			    if(jsd->axis == NULL)
			    {
				jsd->total_axises = 0;
			    }

			    for(i = p; i < jsd->total_axises; i++)
				jsd->axis[i] = NULL;
			}

			/* If axis_num is valid, allocate as needed. */
			if((axis_num >= 0) && (axis_num < jsd->total_axises))
			{
			    if(jsd->axis[axis_num] == NULL)
				jsd->axis[axis_num] = (js_axis_struct *)calloc(
				    1, sizeof(js_axis_struct)
				);
			    axis_ptr = jsd->axis[axis_num];
			}
			/* Failed to get new axis pointer? */
			if(axis_ptr == NULL)
			    axis_num = -1;

                        /* Enter loop to read and handle each line for
			 * this configuration block.
			 */
                        while(1)
                        {
			    DO_GET_LINE

			    /* Minimum */
			    if(!strcasecmp(parm, "Minimum"))
			    {
				if(axis_ptr != NULL)
				    axis_ptr->min = atoi(val);
			    }
                            /* Maximum */
                            else if(!strcasecmp(parm, "Maximum"))
                            {
                                if(axis_ptr != NULL)
                                    axis_ptr->max = atoi(val);
                            }
                            /* Center */
                            else if(!strcasecmp(parm, "Center"))
                            {
                                if(axis_ptr != NULL)
                                    axis_ptr->cen = atoi(val);
                            }
                            /* NullZone */
                            else if(!strcasecmp(parm, "NullZone"))
                            {
                                if(axis_ptr != NULL)
                                    axis_ptr->nz = atoi(val);
                            }
                            /* Tolorance */
                            else if(!strcasecmp(parm, "Tolorance"))
                            {
                                if(axis_ptr != NULL)
				{
                                    axis_ptr->tolorance = atoi(val);
				    axis_ptr->flags |= JSAxisFlagTolorance;
				}
                            }
                            /* Flip */
                            else if(!strcasecmp(parm, "Flip") ||
                                    !strcasecmp(parm, "Flipped")
			    )
                            {
                                if(axis_ptr != NULL)
                                    axis_ptr->flags |= JSAxisFlagFlipped;
                            }
                            /* IsHat */
                            else if(!strcasecmp(parm, "IsHat"))
                            {
                                if(axis_ptr != NULL)
                                    axis_ptr->flags |= JSAxisFlagIsHat;
                            }

                            /* EndAxis */
                            else if(!strcasecmp(parm, "EndAxis"))
                            {
				axis_ptr = NULL;
				axis_num = -1;
                                break;
                            }
			}	/* Read axis block loop. */
                    }
                    /* BeginButton */
                    else if(!strcasecmp(parm, "BeginButton") &&
                            is_this_device
                    )
                    {
			button_ptr = NULL;
                        button_num = atoi(val);
                        if(button_num < 0)
                        {
                            fprintf(     
                                stderr,
                                "%s: Line %i: Invalid button index %i.\n",
                                jsd->calibration_file,
                                lines_read,
                                button_num
                            );
                            button_num = 0;	/* Index must be 0 or higher. */
                        }

                        /* Need to allocate more buttons? */
                        if(button_num >= jsd->total_buttons)
                        {
                            p = jsd->total_buttons;
                            jsd->total_buttons = button_num + 1;
                            jsd->button = (js_button_struct **)realloc(
                                jsd->button,
                                jsd->total_buttons * sizeof(js_button_struct *)
                            );
                            if(jsd->button == NULL)
                            {
                                jsd->total_buttons = 0;
                            }

                            for(i = p; i < jsd->total_buttons; i++)
                                jsd->button[i] = NULL;
                        }

                        /* If button_num is valid, allocate as needed. */
                        if((button_num >= 0) && (button_num < jsd->total_buttons))
                        {
                            if(jsd->button[button_num] == NULL)
                                jsd->button[button_num] = (js_button_struct *)calloc(
                                    1,
                                    sizeof(js_button_struct)
                                );
			    button_ptr = jsd->button[button_num];
                        }
                        /* Failed to get new button pointer? */
                        if(button_ptr == NULL)
                            button_num = -1;

                        /* Enter loop to read and handle each line for
                         * this configuration block.
                         */
                        while(1)
                        {
			    DO_GET_LINE

                            /* EndButton */
                            if(!strcasecmp(parm, "EndButton"))
                            {
				button_ptr = NULL;
				button_num = -1;
                                break;
                            }
			}	/* Read button block loop. */
                    }
                    /* *********************************************** */
                    /* EndJoystick */
                    else if(!strcasecmp(parm, "EndJoystick"))
                    {
			is_this_device = 0;
			break;
                    }
		}	/* Read device block loop. */
	    }
	    /* ******************************************************* */
	    else
	    {
		/* Other parameter, ignore. */

	    }
	}	/* Begin reading file. */

	/* Close file. */
	fclose(fp);

#undef DO_GET_LINE

	return(0);
}
#endif	/* __linux__ */

#ifdef __linux__
/*
 *	Gets a list of calibrated devices found in the specified
 *	calibration file.
 *
 *	The returned list of strings and the pointer array must be
 *	deallocated by the calling function.
 */
char **JSLoadDeviceNamesLinux(
	int *total, const char *calibration
)
{
        FILE *fp;
        int n, lines_read = 0;
        char *strptr, *strptr2;
        char parm[CFG_PARAMETER_MAX];
        char val[CFG_VALUE_MAX];

        int strc = 0;
        char **strv = NULL;

        struct stat stat_buf;


	/* Given calibration file exists? */
        if(calibration == NULL) 
        {
            if(total != NULL)
                (*total) = 0;
            return(NULL);   
        }
        if(stat(calibration, &stat_buf))
        {
            if(total != NULL)
                (*total) = 0;
            return(NULL);
        }

	/* Open calibration file. */
        fp = fopen(calibration, "rb");
        if(fp == NULL)
        {
            if(total != NULL)
                (*total) = 0;
            return(NULL);
        }


/* Reads the next line from fp, first free()ing strptr and then allocating
 * the new line to strptr. Then uses strptr2 to parse the parameter and
 * value, the parsed values will be placed in the variables parm and val,
 * The value of lines_read will also be incremented.
 */
#define DO_GET_LINE     \
{ \
 /* Read next non-comment line. */ \
 free(strptr); \
 strptr = FReadNextLineAllocCount( \
  fp, UNIXCFG_COMMENT_CHAR, &lines_read \
 ); \
 if(strptr == NULL) \
  break; \
 \
 /* Fetch parameter. */ \
 strptr2 = StringCfgParseParm(strptr); \
 if(strptr2 == NULL) \
  continue; \
 strncpy(parm, strptr2, CFG_PARAMETER_MAX); \
 parm[CFG_PARAMETER_MAX - 1] = '\0'; \
 \
 /* Fetch value. */ \
 strptr2 = StringCfgParseValue(strptr); \
 if(strptr2 == NULL) \
  strptr2 = "0"; \
 strncpy(val, strptr2, CFG_VALUE_MAX); \
 val[CFG_VALUE_MAX - 1] = '\0'; \
}

        /* Begin reading calibration file. */
        strptr = NULL;
        while(1)
        {
	    DO_GET_LINE

            /* BeginJoystick */
            if(!strcasecmp(parm, "BeginJoystick"))
            {
                n = strc;
                strc = n + 1;
                strv = (char **)realloc(
                    strv,
                    strc * sizeof(char *)
                );
                if(strv == NULL)
                {
                    strc = 0;
                    continue;
                }

                strv[n] = strdup(val);

                /* Enter loop to read and handle each line for this
                 * configuration block.
                 */
                while(1)
                {
                    DO_GET_LINE

                    /* EndJoystick */
                    if(!strcasecmp(parm, "EndJoystick"))
                    {
                        break;
                    }
                }       /* Read device block loop. */
            }
        }	/* Begin reading calibration file. */

        /* Close file. */
        fclose(fp);
   
#undef DO_GET_LINE

        /* Update total strings returned. */
        if(total != NULL)
            (*total) = strc;

	return(strv);
}

#endif	/* __linux__ */



/*
 *      Initializes the joystick and stores the new initialized values
 *      into the jsd structure.
 *
 *      If the device is not specified (set to NULL), then it will
 *      be defauled to JSDefaultDevice.
 *
 *      If the calibration file is not specified (set to NULL), then
 *      it will be defaulted to JSDefaultCalibration. The HOME
 *      enviroment value will be used as the prefix to the path of
 *      JSDefaultCalibration. The calibration file does not have to
 *      exist.
 *
 *      Available flags are:
 *
 *      JSFlagNonBlocking               Open joystick in non-blocking mode.
 */
int JSInit(
        js_data_struct *jsd,
        const char *device,
        const char *calibration,
	unsigned int flags
)
{
	int i;

#ifdef __linux__
	unsigned char axes = 2;
	unsigned char buttons = 2;
	int version = 0x000800;

# ifndef LINUX_JS_NAME_MAX
#  define LINUX_JS_NAME_MAX	128
# endif

	char name[LINUX_JS_NAME_MAX] = "Unknown";

#endif	/* __linux__ */



	if(jsd == NULL)
	    return(JSBadValue);


	/* Reset values. */
	jsd->axis = NULL;
	jsd->total_axises = 0;

	jsd->button = NULL;
	jsd->total_buttons = 0;

	jsd->device_name = NULL;
	jsd->calibration_file = NULL;

	jsd->fd = -1;
	jsd->flags = 0;
	jsd->driver_version = 0;


	/* Set default device name as needed. */
	if(device == NULL)
	    device = JSDefaultDevice;

	/* Set default calibration file name as needed. */
	if(calibration == NULL)
	{
	    char *strptr1, *strptr2;

	    strptr1 = getenv("HOME");
	    strptr2 = PrefixPaths(
		((strptr1 == NULL) ? "/" : strptr1),
		JSDefaultCalibration
	    );
	    if(strptr2 == NULL)
		strptr2 = JSDefaultCalibration;

	    calibration = strptr2;
	}

	/* Set device and calibration file names. */
	jsd->device_name = StringCopyAlloc(device);
        jsd->calibration_file = StringCopyAlloc(calibration);


#ifdef __linux__
	/* Open joystick. */
	jsd->fd = open(jsd->device_name, O_RDONLY);
	if(jsd->fd < 0)
	{
	    JSClose(jsd);
            return(JSNoAccess);
        }

	/* Fetch device values. */
	/* Raw version string. */
        ioctl(jsd->fd, JSIOCGVERSION, &version);
	jsd->driver_version = (unsigned int)version;

	/* Total number of axises. */
        ioctl(jsd->fd, JSIOCGAXES, &axes);	/* Total axises. */
	jsd->total_axises = axes;

	/* Total number of buttons. */
        ioctl(jsd->fd, JSIOCGBUTTONS, &buttons);	/* Total buttons. */
	jsd->total_buttons = buttons;

	/* Device descriptive name. */
        ioctl(jsd->fd, JSIOCGNAME(LINUX_JS_NAME_MAX), name);
	jsd->name = StringCopyAlloc(name);

	/* Allocate axises. */
	if(jsd->total_axises > 0)
	{
	    jsd->axis = (js_axis_struct **)calloc(
	        jsd->total_axises,
	        sizeof(js_axis_struct *)
	    );
	    if(jsd->axis == NULL)
	    {
	        jsd->total_axises = 0;
	        JSClose(jsd);
	        return(JSNoBuffers);
	    }
	}
	for(i = 0; i < jsd->total_axises; i++)
	{
	    jsd->axis[i] = (js_axis_struct *)calloc(
                1,
                sizeof(js_axis_struct)
            );
            if(jsd->axis == NULL)
            {
                JSClose(jsd);
                return(JSNoBuffers);
            }

	    /* Reset axis values. */
            jsd->axis[i]->cur = JSDefaultCenter;
	    jsd->axis[i]->min = JSDefaultMin;
            jsd->axis[i]->max = JSDefaultMax;
            jsd->axis[i]->cen = JSDefaultCenter;
	    jsd->axis[i]->nz = JSDefaultNullZone;
	    jsd->axis[i]->tolorance = JSDefaultTolorance;
            jsd->axis[i]->flags = 0;
	}

        /* Allocate buttons. */  
	if(jsd->total_buttons > 0)
	{
            jsd->button = (js_button_struct **)calloc(
                jsd->total_buttons,
                sizeof(js_button_struct *)
            );
            if(jsd->button == NULL)   
            {
                jsd->total_buttons = 0;
                JSClose(jsd);   
                return(JSNoBuffers);
	    }
        }
        for(i = 0; i < jsd->total_buttons; i++)
        {
            jsd->button[i] = (js_button_struct *)calloc(
                1,
                sizeof(js_button_struct)
            );
            if(jsd->button == NULL)
            {
                JSClose(jsd);
                return(JSNoBuffers);
            }

	    /* Reset button values. */
            jsd->button[i]->state = JSButtonStateOff;
        }

	/* Set to non-blocking? */
	if(flags & JSFlagNonBlocking)
	{
            fcntl(jsd->fd, F_SETFL, O_NONBLOCK);
	    jsd->flags |= JSFlagNonBlocking;
 	}

        /* Mark successful initialization. */
        jsd->flags |= JSFlagIsInit;

        /* Load calibration from calibration file. */
        JSLoadCalibrationLinux(jsd);

#endif	/* __linux__ */

        /* Set axis tolorance for error correction. */
        JSResetAllAxisTolorance(jsd);


	return(JSSuccess);
}


/*
 *	Updates the information in jsd, returns JSGotEvent if there
 *	was some change or JSNoEvent if there was no change.
 *
 *	jsd needs to be previously initialized by a call to
 *	JSInit().
 */
int JSUpdate(js_data_struct *jsd)
{
	int n;
	int status = JSNoEvent;
#ifdef __linux__
	int keep_handling = 1;
	int bytes_read, cycles;
	struct js_event event;
	js_axis_struct **axis;
	js_button_struct **button;
#endif  /* __linux__ */


        if(jsd == NULL)
            return(status);

	if(jsd->fd < 0)
	    return(status);

#ifdef __linux__
	/* Reset all button state change value on all buttons. */
	for(n = 0, button = jsd->button;
	    n < jsd->total_buttons;
	    n++, button++
	)
	{
	    if(*button != NULL)
		(*button)->changed_state = JSButtonChangedStateNone;
	}

	/* Reset current and previous axis values. */
        for(n = 0, axis = jsd->axis;
            n < jsd->total_axises;
            n++, axis++
        )
        {
            if(*axis != NULL)
                (*axis)->prev = (*axis)->cur;
        }


	/* Handle up to 16 events from joystick driver. */
	cycles = 0;
	while(keep_handling &&
              (cycles < 16)
	)
	{
	    /* Get event. */
	    bytes_read = read(
	        jsd->fd,
	        &event,
	        sizeof(struct js_event)
	    );
	    /* No more events to be read? */
	    if(bytes_read != sizeof(struct js_event))
	        return(status);

	    /* Handle by event type. */
            switch(event.type & ~JS_EVENT_INIT)
	    {
	      /* Axis event. */
	      case JS_EVENT_AXIS:
	        /* Get axis number. */
                n = event.number;

                /* Does axis exist? */
                if(JSIsAxisAllocated(jsd, n))
                {
		    js_axis_struct *axis_ptr = jsd->axis[n];

		    /* Record previous axis position. */
		    axis_ptr->prev = axis_ptr->cur;

		    /* Set new current axis value. */
                    axis_ptr->cur = (int)event.value;

                    /* Record time stamp (in ms). */
                    axis_ptr->last_time = axis_ptr->time;

                    /* Set new time stamp (in ms). */
                    axis_ptr->time = (time_t)event.time;
                }
	        status = JSGotEvent;	/* Mark that we got event. */
	        break;

	      /* Button event. */
              case JS_EVENT_BUTTON:
	        /* Get button number. */
	        n = event.number;

	        /* Does button exist? */
                if(JSIsButtonAllocated(jsd, n))
	        {
		    js_button_struct *button_ptr = jsd->button[n];

		    /* Record previous state. */
		    button_ptr->prev_state = button_ptr->state;

		    /* Set new button state. */
                    button_ptr->state = ((event.value) ?
		        JSButtonStateOn : JSButtonStateOff
		    );

		    /* Update state change. */
		    if((button_ptr->prev_state == JSButtonStateOn) &&
                       (button_ptr->state == JSButtonStateOff)
		    )
			button_ptr->changed_state = JSButtonChangedStateOnToOff;
                    else if((button_ptr->prev_state == JSButtonStateOff) &&
                            (button_ptr->state == JSButtonStateOn)
                    )
			button_ptr->changed_state = JSButtonChangedStateOffToOn;

                    /* Record time stamp (in ms). */
		    button_ptr->last_time = button_ptr->time;

		    /* Set new time stamp (in ms). */
                    button_ptr->time = (time_t)event.time;
	        }
	        status = JSGotEvent;	/* Mark that we got event. */
	        break;

	      /* Other event. */
	      default:
		keep_handling = 0;	/* Stop handling events. */
		break;
	    }

	    cycles++;
	}
#endif	/* __linux__ */


        return(status);
}

/*
 *	Closes the joystick and deallocates all resources.
 */
void JSClose(js_data_struct *jsd)
{
	int i;

	if(jsd == NULL)
	    return;

	/* Close descriptor to joystick. */
	if(jsd->fd > -1)
	{
	    close(jsd->fd);
	    jsd->fd = -1;
	}

	/* Free name. */
	free(jsd->name);
	jsd->name = NULL;


	/* Free all axises. */
	for(i = 0; i < jsd->total_axises; i++)
	    free(jsd->axis[i]);
	free(jsd->axis);
	jsd->axis = NULL;

	jsd->total_axises = 0;

	/* Free all buttons. */
	for(i = 0; i < jsd->total_buttons; i++)
            free(jsd->button[i]);
        free(jsd->button);
        jsd->button = NULL;
        jsd->total_buttons = 0;


	/* Free device name. */
	free(jsd->device_name);
	jsd->device_name = NULL;

	/* Free calibration file name. */
	free(jsd->calibration_file);
	jsd->calibration_file = NULL;



	jsd->flags = 0;
	jsd->driver_version = 0;


	return;
}


/*
 *	Dummy function, for ANSI C compliancy.
 */
int JSMain(int argc, char *argv[])
{
	int x;

	x = 0;

	return(x);
}
